import pyarrow  # type: ignore
from pyarrow.cffi import ffi
import rpy2.rinterface as rinterface
import rpy2.robjects as robjects
import rpy2.robjects.conversion as conversion
import rpy2.robjects.packages as packages
import typing
import warnings


rarrow = packages.importr('arrow')

if rinterface.BoolSexpVector(
        rinterface.evalr("""
        utils::packageVersion("arrow") < base::package_version("12.0")
        """)
)[0]:
    warnings.warn('Segfaults may occur with the R package arrow < 12.0')

# make sure a version is installed with the C API
_rarrow_has_c_api = rinterface.BoolSexpVector(
    rinterface.evalr("""
    utils::packageVersion("arrow") >= base::package_version("5.0.0")
    """)
)[0]

if not _rarrow_has_c_api:
    raise ValueError("rpy2_arrow requires R 'arrow' package version >= 5.0.0")

# In arrow >= 7.0.0, pointers can be passed as externalptr,
# bit64::integer64(), or string, all of which prevent possible
# problems with the previous versions which required a double().
_use_r_ptr_string = rinterface.BoolSexpVector(
    rinterface.evalr("""
    utils::packageVersion("arrow") >= base::package_version("6.0.1.9000")
    """)
)[0]


def _rarrow_ptr(ptr) -> typing.Union[str, float]:
    ptr_value = int(ffi.cast('uintptr_t', ptr))
    return str(ptr_value) if _use_r_ptr_string else float(ptr_value)


def _c_ptr_to_int(ptr) -> int:
    return int(ffi.cast('uintptr_t', ptr))


def pyarrow_to_r_array(
        obj: 'pyarrow.lib.Array'
):
    """Create an R `arrow::Arrow` object from a pyarrow Array.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """
    array_ptr = ffi.new('struct ArrowArray*')
    schema_ptr = ffi.new('struct ArrowSchema*')

    obj._export_to_c(_c_ptr_to_int(array_ptr), _c_ptr_to_int(schema_ptr))
    return rarrow.Array['import_from_c'](_rarrow_ptr(array_ptr), _rarrow_ptr(schema_ptr))


def rarrow_to_py_array(
        obj: robjects.Environment
) -> pyarrow.Array:
    """Create a pyarrow array from an R `arrow::Array` object.

    This is sharing the C/C++ object between the two languages.
    """
    array_ptr = ffi.new('struct ArrowArray*')
    schema_ptr = ffi.new('struct ArrowSchema*')

    obj['export_to_c'](_rarrow_ptr(array_ptr), _rarrow_ptr(schema_ptr))
    return pyarrow.lib.Array._import_from_c(_c_ptr_to_int(array_ptr), _c_ptr_to_int(schema_ptr))


def pyarrow_to_r_recordbatch(
        obj: 'pyarrow.lib.RecordBatch'
):
    """Create an R `arrow::RecordBatch` object from a pyarrow RecordBatch.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """
    array_ptr = ffi.new('struct ArrowArray*')
    schema_ptr = ffi.new('struct ArrowSchema*')

    obj._export_to_c(_c_ptr_to_int(array_ptr), _c_ptr_to_int(schema_ptr))
    return rarrow.RecordBatch['import_from_c'](_rarrow_ptr(array_ptr), _rarrow_ptr(schema_ptr))


def rarrow_to_py_recordbatch(
        obj: robjects.Environment
) -> pyarrow.lib.RecordBatch:
    """Create a pyarrow record batch from an R `arrow::Array` object.

    This is sharing the C/C++ object between the two languages.
    """
    array_ptr = ffi.new('struct ArrowArray*')
    schema_ptr = ffi.new('struct ArrowSchema*')

    obj['export_to_c'](_rarrow_ptr(array_ptr), _rarrow_ptr(schema_ptr))
    return pyarrow.lib.RecordBatch._import_from_c(_c_ptr_to_int(array_ptr), _c_ptr_to_int(schema_ptr))


def pyarrow_to_r_recordbatchreader(
        obj: 'pyarrow.lib.RecordBatchReader'
):
    """Create an R `arrow::RecordBatchReader` from a pyarrow RecordBatchReader.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    stream_ptr = ffi.new('struct ArrowArrayStream*')
    obj._export_to_c(_c_ptr_to_int(stream_ptr))
    return rarrow.RecordBatchReader['import_from_c'](_rarrow_ptr(stream_ptr))


def rarrow_to_py_recordbatchreader(
        obj: robjects.Environment
) -> pyarrow.lib.RecordBatchReader:
    """Create a pyarrow RecordBatchReader from an R `arrow::RecordBatchReader` object.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    stream_ptr = ffi.new('struct ArrowArrayStream*')
    obj['export_to_c'](_rarrow_ptr(stream_ptr))
    return pyarrow.lib.RecordBatchReader._import_from_c(_c_ptr_to_int(stream_ptr))


def pyarrow_to_r_chunkedarray(
        obj: 'pyarrow.lib.ChunkedArray'
):
    """Create an R `arrow::ChunkedArray` object from a pyarrow ChunkedArray.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """
    chunks = tuple(pyarrow_to_r_array(x) for x in obj.chunks)
    res = rarrow.ChunkedArray['create'](*chunks)
    return res


def rarrow_to_py_chunkedarray(
        obj: robjects.Environment
) -> pyarrow.lib.ChunkedArray:
    """Create a pyarrow chunked array from an R `arrow::ChunkedArray` object.

    This is sharing the C/C++ object between the two languages.
    """
    chunks = tuple(rarrow_to_py_array(x) for x in obj['chunks'])
    return pyarrow.chunked_array(chunks)


def pyarrow_to_r_datatype(
        obj: 'pyarrow.lib.DataType'
):
    schema_ptr = ffi.new('struct ArrowSchema*')
    obj._export_to_c(_c_ptr_to_int(schema_ptr))
    return rarrow.DataType['import_from_c'](_rarrow_ptr(schema_ptr))


def rarrow_to_py_datatype(
        obj: robjects.Environment
) -> pyarrow.lib.DataType:
    """Create a pyarrow.lib.DataType from an R `arrow::DataType` object.

    This is sharing the C/C++ object between the two languages.
    """
    schema_ptr = ffi.new('struct ArrowSchema*')
    obj['export_to_c'](_rarrow_ptr(schema_ptr))
    return pyarrow.lib.DataType._import_from_c(_c_ptr_to_int(schema_ptr))


def pyarrow_to_r_field(
        obj: 'pyarrow.lib.Field'
):
    schema_ptr = ffi.new('struct ArrowSchema*')
    obj._export_to_c(_c_ptr_to_int(schema_ptr))
    return rarrow.Field['import_from_c'](_rarrow_ptr(schema_ptr))


def rarrow_to_py_field(
        obj: robjects.Environment
) -> pyarrow.lib.Field:
    """Create a pyarrow.lib.Field from an R `arrow::DataType` object.

    This is sharing the C/C++ object between the two languages.
    """

    schema_ptr = ffi.new('struct ArrowSchema*')
    obj['export_to_c'](_rarrow_ptr(schema_ptr))
    return pyarrow.lib.Field._import_from_c(_c_ptr_to_int(schema_ptr))


_as_arrow_table_from_stream_ptr = rinterface.SexpClosure(
    rinterface.evalr(
        """
        function(recordbatchreader) {
        arrow::as_arrow_table(recordbatchreader)
        }
        """
    )
)


def _pyarrow_table_to_r_table_ri(tbl: pyarrow.lib.Table) -> rinterface.SexpEnvironment:
    """Create an R `arrow` Table to mirror an Arrow Table in `pyarrow`."""
    recordbatchreader = pyarrow_to_r_recordbatchreader(tbl.to_reader())
    return _as_arrow_table_from_stream_ptr(recordbatchreader)


def pyarrow_table_to_r_table(obj: pyarrow.lib.Table):
    """Create an R `arrow::Table` object from a pyarrow Table.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.

    :param:obj: A :class:`pyarrow.lib.Table`.
    :return: The result of conversion rules for R `arrow` `Table` objects.
    By default a :class:`rpy2.robjects.Environment`.
    """
    # `res` is a low-level (rinterface-level) rpy2 object. This is an
    # rpy2 robject-level function. Use the conversion.
    res = _pyarrow_table_to_r_table_ri(obj)
    converter = (
        conversion.get_conversion()  # rpy2 >=3.5.2
        if hasattr(conversion, 'get_conversion') else
        conversion.converter  # rpy2 <3.5.2
    )
    return converter.py2rpy(res)


def rarrow_to_py_table(
        obj: robjects.Environment,
        rpy2py: typing.Optional[
            conversion.Converter] = None
) -> pyarrow.Table:
    """Create a pyarrow Table from an R `arrow::Table` object.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2.
    """

    if rpy2py is None:
        rpy2py = converter
    # TODO: rpy2 conversion forces something a little kludgy here.
    columns = [
        (rpy2py._rpy2py_nc_map[rinterface.SexpEnvironment][x.rclass[0]](x))
        for x in obj['columns']
    ]
    schema = rarrow_to_py_schema(obj['schema'])
    py_table = pyarrow.Table.from_arrays(columns,
                                         schema=schema)
    return py_table


def pyarrow_to_r_schema(
        obj: 'pyarrow.lib.Schema'
):
    """Create an R `arrow::Schema` object from a pyarrow Schema.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    schema_ptr = ffi.new('struct ArrowSchema*')
    obj._export_to_c(_c_ptr_to_int(schema_ptr))
    return rarrow.Schema['import_from_c'](_rarrow_ptr(schema_ptr))


def rarrow_to_py_schema(
        obj: robjects.Environment
) -> pyarrow.Schema:
    """Create a pyarrow Schema from an R `arrow::Schema` object.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    schema_ptr = ffi.new('struct ArrowSchema*')
    obj['export_to_c'](_rarrow_ptr(schema_ptr))
    return pyarrow.lib.Schema._import_from_c(_c_ptr_to_int(schema_ptr))


converter: conversion.Converter = conversion.Converter(
    'default arrow conversion',
    template=robjects.default_converter
)

# Pyarrow to R arrow.
converter.py2rpy.register(pyarrow.lib.Array, pyarrow_to_r_array)
converter.py2rpy.register(pyarrow.lib.Field, pyarrow_to_r_field)
converter.py2rpy.register(pyarrow.lib.ChunkedArray,
                          pyarrow_to_r_chunkedarray)
converter.py2rpy.register(pyarrow.lib.RecordBatch,
                          pyarrow_to_r_recordbatch)
converter.py2rpy.register(pyarrow.lib.RecordBatchReader,
                          pyarrow_to_r_recordbatchreader)
converter.py2rpy.register(pyarrow.lib.Schema, pyarrow_to_r_schema)
converter.py2rpy.register(pyarrow.lib.Table, pyarrow_table_to_r_table)
converter.py2rpy.register(pyarrow.lib.DataType, pyarrow_to_r_datatype)

# R arrow to pyarrow.
converter._rpy2py_nc_map.update(
    {
        rinterface.SexpEnvironment:
        conversion.NameClassMap(robjects.Environment)
    }
)

# TODO: use complete class name hierarchy to be safer?
converter._rpy2py_nc_map[rinterface.SexpEnvironment].update(
    {
        'Array': rarrow_to_py_array,
        'ChunkedArray': rarrow_to_py_chunkedarray,
        'Field': rarrow_to_py_field,
        'RecordBatch': rarrow_to_py_recordbatch,
        'RecordBatchReader': rarrow_to_py_recordbatchreader,
        'Schema': rarrow_to_py_schema,
        'Table': rarrow_to_py_table,
        'Type': rarrow_to_py_datatype
    }
)



import polars
import rpy2.rinterface
import rpy2.robjects
import rpy2.robjects.conversion as conversion
import rpy2_arrow.arrow as rpy2arrow
import types
import typing


# This can be accessed as a global to lazily import
# the R package polars.
rpack_polars: typing.Optional[types.ModuleType] = None
rpack_arrow: typing.Optional[types.ModuleType] = None


def ensure_r_polars() -> types.ModuleType:
    global rpack_polars
    if rpack_polars is None:
        rpack_polars = rpy2.robjects.packages.importr('polars',
                                                      on_conflict='warn')
    rpack_polars_version_info = rpack_polars.__version__.split('.')
    assert tuple(int(_) for _ in rpack_polars_version_info[:2]) >= (0, 12)
    return rpack_polars


def ensure_r_arrow() -> types.ModuleType:
    global rpack_arrow
    if rpack_arrow is None:
        rpack_arrow = rpy2.robjects.packages.importr('arrow',
                                                     on_conflict='warn')
    return rpack_arrow


def pypolars_to_rarrow_dataframe(
        dataf: polars.DataFrame
) -> rpy2.robjects.Environment:
    _ = dataf.to_arrow()
    return rpy2arrow.pyarrow_table_to_r_table(_)


def rarrow_to_pypolars_dataframe(
        dataf: rpy2.robjects.Environment
) -> polars.DataFrame:
    _ = rpy2arrow.rarrow_to_py_table(dataf)
    res = polars.from_arrow(_)
    if not isinstance(res, polars.DataFrame):
        raise ValueError(
            'polars.from_arrow() did not convert the object to a '
            'polars.DataFrame'
        )
    return res


def pypolars_to_rpolars_dataframe(
        dataf: polars.DataFrame
) -> rpy2.robjects.ExternalPointer:
    r_arrow_table = pypolars_to_rarrow_dataframe(dataf)
    rpack_polars = ensure_r_polars()
    # TODO: There appear to be an odd shortcircuiting that requires toggling
    # additional conversion off.
    with rpy2.robjects.default_converter.context():
        return rpack_polars.as_polars_df(r_arrow_table)


# TODO: rpy2.rinterface.SexpExtPtr should have an robjects-level wrapper?
def rpolar_to_pypolars_dataframe(
        dataf: rpy2.rinterface.SexpExtPtr
) -> polars.DataFrame:
    # R polars to R arrow.
    rpack_arrow = ensure_r_arrow()
    ensure_r_polars()
    r_arrow_table = rpack_arrow.as_arrow_table(dataf)
    return rarrow_to_pypolars_dataframe(r_arrow_table)


converter: conversion.Converter = conversion.Converter(
    'default polars conversion',
    template=rpy2.robjects.default_converter
)

converter.py2rpy.register(polars.dataframe.frame.DataFrame,
                          pypolars_to_rpolars_dataframe)

converter._rpy2py_nc_map.update(
    {
        rpy2.rinterface.SexpEnvironment:
        conversion.NameClassMap(rpy2.robjects.Environment),
        rpy2.rinterface.SexpExtPtr:
        conversion.NameClassMap(rpy2.robjects.ExternalPointer)
    }
)

converter._rpy2py_nc_map[rpy2.rinterface.SexpEnvironment].update(
    {
        'Table': rarrow_to_pypolars_dataframe,
    }
)

converter._rpy2py_nc_map[rpy2.rinterface.SexpExtPtr].update(
    {
        'RPolarsDataFrame': rpolar_to_pypolars_dataframe,
    }
)


def pl_to_rpl(df):
    """Convenience shortcut to convert a polars object to a R polars object."""
    with converter.context() as conversion_ctx:
        return conversion_ctx.py2rpy(df)


def rpl_to_pl(df):
    """Convenience shortcut to convert a R polars object to a polars object."""
    with converter.context() as conversion_ctx:
        return conversion_ctx.rpy2py(df)



try:
    import polars
    import rpy2_arrow.polars as rpy2polars
    HAS_POLARS = True
except ImportError:
    HAS_POLARS = False

import pytest
import rpy2.rinterface
import rpy2.robjects

R_DOLLAR = rpy2.rinterface.baseenv['$']
R_ASVECTOR = rpy2.rinterface.baseenv['as.vector']
R_SQBRACKET = rpy2.rinterface.baseenv['[']
R_LENGTH = rpy2.rinterface.baseenv['length']
R_EQUAL = rpy2.rinterface.baseenv['==']


def _cmp_simple(v1, v2):
    return tuple(v1) == tuple(v2)


def _cmp_float(v1, v2, tol=1E-5):
    return all(abs(x - y) < tol for x, y in zip(v1, v2))


@pytest.mark.skipif(not HAS_POLARS,
                    reason='The Python package "polars" is required.')
class TestPolars:
    def test_ensure_r_polars(self):
        rpack_polars = rpy2polars.ensure_r_polars()
        assert rpack_polars.__name__ == 'polars'

    def test_ensure_r_arrow(self):
        rpack_arrow = rpy2polars.ensure_r_arrow()
        assert rpack_arrow.__rname__ == 'arrow'

    @pytest.mark.parametrize(
        'values,dtype,rarrowclass,cmp',
        [
            ([1, 2], polars.Int8, 'Int8', _cmp_simple),
            ([1, 2], polars.Int16, 'Int16', _cmp_simple),
            ([1, 2], polars.Int32, 'Int32', _cmp_simple),
            ([1, 2], polars.Int64, 'Int64', _cmp_simple),
            ([1.1, 2.1], polars.Float32, 'Float32', _cmp_float),
            ([1.1, 2.1], polars.Float64, 'Float64', _cmp_float),
            (['wx', 'yz'], polars.Utf8, 'LargeUtf8', _cmp_simple),
            (['wx', 'yz', 'wx'], polars.Categorical,
             'DictionaryType', _cmp_simple)
        ])
    def test_pypolars_to_rarrow_dataframe(
            self, values, dtype, rarrowclass, cmp
    ):
        podataf = polars.DataFrame({'a': values}, schema={'a': dtype})
        rar_table = rpy2polars.pypolars_to_rarrow_dataframe(podataf)
        assert rar_table.rclass[0] == 'Table'
        assert tuple(
            R_DOLLAR(R_DOLLAR(rar_table, 'schema'), 'names')
        ) == ('a', )
        field = R_SQBRACKET(
            R_DOLLAR(R_DOLLAR(rar_table, 'schema'), 'fields'),
            1
        )[0]
        assert R_DOLLAR(field, 'type').rclass[0] == rarrowclass

    @pytest.mark.parametrize(
        'values,rstr,cmp',
        [
            ([int(1), int(2)], 'as.integer(c(1, 2))', _cmp_simple),
            ([1.1, 2.1], 'c(1.1, 2.1)', _cmp_float),
            (['wx', 'yz'], 'c("wx", "yz")', _cmp_simple),
            (['wx', 'yz', 'wx'], 'factor(c("wx", "yz", "wx"))', _cmp_simple),
        ])
    def test_rarrow_to_pypolars_dataframe(self, values, rstr, cmp):
        rartable = rpy2.robjects.r(
            f'arrow::arrow_table(data.frame(a = {rstr}))'
        )
        podataf = rpy2polars.rarrow_to_pypolars_dataframe(rartable)
        assert cmp(podataf['a'], values)

    def test_rpolar_to_pypolars_dataframe(self):
        rpack_polars = rpy2polars.ensure_r_polars()
        rpodataf = rpack_polars.pl['DataFrame'](
            a=rpy2.robjects.IntVector([1, 2]),
            b=rpy2.robjects.StrVector(['wx', 'yz'])
        )
        podataf = rpy2polars.rpolar_to_pypolars_dataframe(rpodataf)
        assert tuple(podataf['a']) == (1, 2)
        assert tuple(podataf['b']) == ('wx', 'yz')

    @pytest.mark.parametrize(
        'values,dtype,rpotype,cmp',
        [
            ([1, 2], polars.Int8, 'Int8', _cmp_simple),
            ([1, 2], polars.Int16, 'Int16', _cmp_simple),
            ([1, 2], polars.Int32, 'Int32', _cmp_simple),
            ([1, 2], polars.Int64, 'Int64', _cmp_simple),  # Fails.
            ([1.1, 2.1], polars.Float32, 'Float32', _cmp_float),
            ([1.1, 2.1], polars.Float64, 'Float64', _cmp_float),
            (['wx', 'yz'], polars.Utf8, 'Utf8', _cmp_simple),
            (['wx', 'yz', 'wx'], polars.Categorical,
             'Categorical', _cmp_simple)
        ])
    def test_converter_py2rpy_dataframe(self, values, dtype, rpotype, cmp):
        podataf = polars.DataFrame({'a': values}, schema={'a': dtype})
        globalenv = rpy2.robjects.globalenv
        with rpy2polars.converter.context():
            globalenv['podataf'] = podataf
        r_podataf = globalenv['podataf']
        assert tuple(r_podataf.rclass) == ('RPolarsDataFrame',)

        assert tuple(
            R_DOLLAR(r_podataf, 'schema').names
        ) == ('a', )
        field = R_SQBRACKET(
            R_DOLLAR(r_podataf, 'schema'), 1
        )[0]
        assert R_EQUAL(
            field,
            R_DOLLAR(getattr(rpy2polars.rpack_polars, 'pl'), rpotype)
        )

    @pytest.mark.parametrize(
        'values,rstr,cmp,dtype',
        [
            ([int(1), int(2)], 'as.integer(c(1, 2))',
             _cmp_simple, polars.Int32),
            ([1.1, 2.1], 'c(1.1, 2.1)', _cmp_float, polars.Float64),
            (['wx', 'yz'], 'c("wx", "yz")', _cmp_simple, polars.Utf8),
            (['wx', 'yz', 'wx'], 'factor(c("wx", "yz", "wx"))',
             _cmp_simple, polars.Categorical),
        ])
    def test_converter_rpy2py_dataframe(self, values, rstr, cmp, dtype):
        rpy2.robjects.r(
            f'require(polars); podataf <- pl$DataFrame(a = {rstr})'
        )
        podataf_ri = rpy2.rinterface.globalenv['podataf']
        with rpy2polars.converter.context() as ctx:
            podataf = ctx.rpy2py(podataf_ri)
        assert isinstance(podataf, polars.dataframe.frame.DataFrame)
        assert cmp(podataf['a'], values)
        assert podataf['a'].dtype == dtype

    @pytest.mark.parametrize(
        'rstr,cls',
        [
            ('pl$DataFrame(a = as.integer(c(1, 2)))',
             polars.DataFrame),
            ('pl$DataFrame(a = as.integer(c(1, 2)), b = c("a", "b"))',
             polars.DataFrame)

        ]
    )
    def test_rpl_to_pl(self, rstr, cls):
        rplobj_ri = rpy2.rinterface.evalr(
            f'require(polars); {rstr}'
        )
        plobj = rpy2polars.rpl_to_pl(rplobj_ri)
        assert isinstance(plobj, cls)

    def test_pl_to_rpl(self):
        plobj = polars.DataFrame({'a': [1, 2, 3]})
        cls = rpy2.robjects.ExternalPointer
        rplobj = rpy2polars.pl_to_rpl(plobj)
        assert isinstance(rplobj, cls)



import rpy2_r6.r6b as r6b
import rpy2_arrow.arrow as pyra
import rpy2.rinterface as rinterface
import rpy2.robjects
import rpy2.robjects.conversion

# Python proxies for the R6 class factories
array_factory = r6b.R6DynamicClassGenerator(pyra.rarrow.Array)
recordbatch_factory = r6b.R6DynamicClassGenerator(pyra.rarrow.RecordBatch)
chunkedarray_factory = r6b.R6DynamicClassGenerator(pyra.rarrow.ChunkedArray)
schema_factory = r6b.R6DynamicClassGenerator(pyra.rarrow.Schema)
table_factory = r6b.R6DynamicClassGenerator(pyra.rarrow.Table)


# Conversion functions and rules
converter = rpy2.robjects.conversion.Converter(
    'R6b conversion for pyarrow/arrow',
    template=rpy2.robjects.default_converter
)


def rpy2py_array(obj):
    return array_factory.__R6CLASS__(obj)


def rpy2py_recordbatch(obj):
    return recordbatch_factory.__R6CLASS__(obj)


def rpy2py_chunkedarray(obj):
    return chunkedarray_factory.__R6CLASS__(obj)


def rpy2py_schema(obj):
    return schema_factory.__R6CLASS__(obj)


def rpy2py_table(obj):
    return table_factory.__R6CLASS__(obj)


(converter.rpy2py_nc_name[rinterface.SexpEnvironment]
 .update({
     'Array': array_factory.__R6CLASS__,
     'ChunkedArray': chunkedarray_factory.__R6CLASS__,
     'RecordBatch': recordbatch_factory.__R6CLASS__,
     'Table': table_factory.__R6CLASS__,
     'Schema': schema_factory.__R6CLASS__
 }))



import array
import pandas
import rpy2.rinterface as rinterface
import rpy2_arrow.arrow as pyra
import pyarrow


def test_py2r_Array():
    a = array.array('i', [1, 2, 3])
    py_ar = pyarrow.array(a)
    r_ar = pyra.pyarrow_to_r_array(py_ar)
    assert isinstance(r_ar, rinterface.SexpEnvironment)


def test_r2py_Array():
    r_ar = rinterface.evalr("""
    require('arrow')
    arrow::Array$create(c(1, 2, 3))
    """)
    py_ar = pyra.rarrow_to_py_array(r_ar)
    assert isinstance(py_ar, pyarrow.Array)


def test_py2r_DataType():
    py_dt = pyarrow.bool_()
    r_dt = pyra.pyarrow_to_r_datatype(py_dt)
    assert isinstance(r_dt, rinterface.SexpEnvironment)


def test_r2py_DataType():
    r_dt = rinterface.evalr("""
    require('arrow')
    arrow::bool()
    """)
    py_dt = pyra.rarrow_to_py_datatype(r_dt)
    assert isinstance(py_dt, pyarrow.DataType)


def test_py2r_Field():
    py_fd = pyarrow.field('foo', pyarrow.bool_())
    r_fd = pyra.pyarrow_to_r_field(py_fd)
    assert isinstance(r_fd, rinterface.SexpEnvironment)


def test_r2py_Field():
    r_fd = rinterface.evalr("""
    require('arrow')
    arrow::Field$create("foo", arrow::bool())
    """)
    py_fd = pyra.rarrow_to_py_field(r_fd)
    assert isinstance(py_fd, pyarrow.lib.Field)


def test_py2r_RecordBatch():
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_rb = pyarrow.record_batch(dataf)
    r_rb = pyra.pyarrow_to_r_recordbatch(py_rb)
    assert isinstance(r_rb, rinterface.SexpEnvironment)


def test_r2py_RecordBatch():
    r_rb = rinterface.evalr("""
    require('arrow')
    arrow::RecordBatch$create(a = c(1L, 2L, 3L), b = c(4L, 5L, 6L))
    """)
    py_rb = pyra.rarrow_to_py_recordbatch(r_rb)
    assert isinstance(py_rb, pyarrow.RecordBatch)


def test_py2r_RecordBatchReader():
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_tb = pyarrow.table(dataf)
    py_rbr = pyarrow.lib.RecordBatchReader.from_batches(
        py_tb.schema,
        py_tb.to_batches())
    r_rbr = pyra.pyarrow_to_r_recordbatchreader(py_rbr)
    assert isinstance(r_rbr, rinterface.SexpEnvironment)


def test_r2py_RecordBatchReader():
    # An R-native way to get a RecordBatchReader segfaults under CRAN
    # arrow but not under development arrow, so use the exported
    # Python RecordBatchReader instead.
    # r_rbr = rinterface.evalr("""
    # require('arrow')
    # tb <- arrow::Table$create(a = c(1L, 2L, 3L), b = c(4L, 5L, 6L))
    # scanner <- Scanner$create(tb)
    # scanner$ToRecordBatchReader()
    # """)
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_tb = pyarrow.table(dataf)
    py_rbr = pyarrow.lib.RecordBatchReader.from_batches(
        py_tb.schema,
        py_tb.to_batches())
    r_rbr = pyra.pyarrow_to_r_recordbatchreader(py_rbr)
    assert isinstance(r_rbr, rinterface.SexpEnvironment)

    # convert back to Python
    py_rbr = pyra.rarrow_to_py_recordbatchreader(r_rbr)
    assert isinstance(py_rbr, pyarrow.lib.RecordBatchReader)


def test_py2r_ChunkedArray():
    py_ca = pyarrow.chunked_array([[1, 2, 3], [4, 5]])
    r_ca = pyra.pyarrow_to_r_chunkedarray(py_ca)
    assert isinstance(r_ca, rinterface.SexpEnvironment)


def test_r2py_ChunkedArray():
    r_ca = rinterface.evalr("""
    require('arrow')
    arrow::ChunkedArray$create(c(1, 2, 3))
    """)
    py_ca = pyra.rarrow_to_py_chunkedarray(r_ca)
    assert isinstance(py_ca, pyarrow.ChunkedArray)


def test_py2r_Table():
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_tb = pyarrow.Table.from_pandas(dataf)
    r_tb = pyra.pyarrow_table_to_r_table(py_tb)
    assert isinstance(r_tb, rinterface.SexpEnvironment)


def test_r2py_Table():
    r_tb = rinterface.evalr("""
    require('arrow')
    arrow::Table$create(
      data.frame(a = c(1, 2, 3),
                 b = c(4, 5, 6))
    )
    """)
    py_tb = pyra.rarrow_to_py_table(r_tb)
    assert isinstance(py_tb, pyarrow.Table)



import rpy2.rinterface as rinterface
import rpy2.robjects.conversion as conversion
import rpy2.robjects as ro
import rpy2_arrow.arrow as pyra
import rpy2_arrow.r6b as r6b_ar
import rpy2_r6.r6b as r6b
import pyarrow
import pytest


@pytest.mark.xfail
@pytest.mark.parametrize(
    'pyarrow_constructor,pyr_py2r',
    [
        (pyarrow.array, pyra.pyarrow_to_r_array),
        (pyarrow.chunked_array, pyra.pyarrow_to_r_chunkedarray)
    ]
)
def test_rpy2py_arrays(pyarrow_constructor, pyr_py2r):
    a = [1, 2, 3]
    py_ar = pyarrow_constructor(a)
    r6_ar = pyr_py2r(py_ar)
    assert isinstance(r6_ar, rinterface.SexpEnvironment)
    # TODO: bug in rpy2's layering of conversion rules?
    # with conversion.local_converter(
    #         ro.default_converter + r6b_ar.converter
    # ).context() as cv:
    env_map = (ro.conversion
               .converter.rpy2py_nc_name[rinterface.SexpEnvironment])
    with conversion.NameClassMapContext(
            env_map,
            r6b_ar.converter.rpy2py_nc_name[rinterface.SexpEnvironment]._map
    ).context():
        r6_ar = pyr_py2r(py_ar)
        assert isinstance(r6_ar, r6b.R6)
